from unittest import TestCase

from flake8_quotes import QuoteChecker
from test.test_checks import get_absolute_path


class DocstringTestChecks(TestCase):
    def test_require_double_docstring_double_present(self):
        class Options():
            inline_quotes = 'single'
            multiline_quotes = 'single'
            docstring_quotes = 'double'
        QuoteChecker.parse_options(Options)

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 5, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 16, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 20, 'line': 21, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 8, 'line': 30, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 12, 'line': 35, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_module_multiline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 4, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 0, 'line': 9, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_module_singleline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 2, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 0, 'line': 6, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_class.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 3, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 22, 'line': 5, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_function.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 3, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 11, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 38, 'line': 15, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 17, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 21, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

    def test_require_single_docstring_double_present(self):
        class Options():
            inline_quotes = 'single'
            multiline_quotes = 'double'
            docstring_quotes = 'single'
        QuoteChecker.parse_options(Options)

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 4, 'line': 12, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 8, 'line': 24, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_module_multiline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_module_singleline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_class.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 2, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 8, 'line': 6, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 28, 'line': 9, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_doubles_function.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 2, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 4, 'line': 8, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

    def test_require_double_docstring_single_present(self):
        class Options():
            inline_quotes = 'single'
            multiline_quotes = 'single'
            docstring_quotes = 'double'
        QuoteChecker.parse_options(Options)

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 4, 'line': 14, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 8, 'line': 26, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_module_multiline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_module_singleline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 1, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_class.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 2, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 8, 'line': 6, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 28, 'line': 9, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_function.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 2, 'message': 'Q002 Remove bad quotes from docstring'},
            {'col': 4, 'line': 8, 'message': 'Q002 Remove bad quotes from docstring'},
        ])

    def test_require_single_docstring_single_present(self):
        class Options():
            inline_quotes = 'single'
            multiline_quotes = 'double'
            docstring_quotes = 'single'
        QuoteChecker.parse_options(Options)

        multiline_checker = QuoteChecker(None, filename=get_absolute_path('data/docstring_singles.py'))
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 5, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 20, 'line': 11, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 18, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 20, 'line': 23, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 8, 'line': 32, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 12, 'line': 37, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_module_multiline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 4, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 0, 'line': 9, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_module_singleline.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 0, 'line': 2, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 0, 'line': 6, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_class.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 3, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 22, 'line': 5, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])

        multiline_checker = QuoteChecker(
            None,
            filename=get_absolute_path('data/docstring_singles_function.py')
        )
        self.assertEqual(list(multiline_checker.get_quotes_errors(multiline_checker.get_file_contents())), [
            {'col': 4, 'line': 3, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 11, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 38, 'line': 15, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 17, 'message': 'Q001 Remove bad quotes from multiline string'},
            {'col': 4, 'line': 21, 'message': 'Q001 Remove bad quotes from multiline string'},
        ])



import tokenize
from unittest import TestCase

from flake8_quotes import Token, get_docstring_tokens
from test.test_checks import get_absolute_path


class GetDocstringTokensTests(TestCase):
    def _get_docstring_tokens(self, filename):
        with open(get_absolute_path(filename), 'r') as f:
            tokens = [Token(t) for t in tokenize.generate_tokens(f.readline)]
        return get_docstring_tokens(tokens)

    def test_get_docstring_tokens_absent(self):
        self.assertEqual(self._get_docstring_tokens('data/doubles.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/doubles_multiline_string.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/doubles_noqa.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/doubles_wrapped.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/multiline_string.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/no_qa.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/singles.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/singles_multiline_string.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/singles_noqa.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/singles_wrapped.py'), set())
        self.assertEqual(self._get_docstring_tokens('data/docstring_not_docstrings.py'), set())

    def test_get_docstring_tokens_doubles(self):
        with open(get_absolute_path('data/docstring_doubles.py'), 'r') as f:
            tokens = [Token(t) for t in tokenize.generate_tokens(f.readline)]
        docstring_tokens = {t.string for t in get_docstring_tokens(tokens)}
        self.assertEqual(docstring_tokens, {
            '"""\nDouble quotes multiline module docstring\n"""',
            '"""\n    Double quotes multiline class docstring\n    """',
            '"""\n        Double quotes multiline function docstring\n        """',
        })

    def test_get_docstring_tokens_singles(self):
        with open(get_absolute_path('data/docstring_singles.py'), 'r') as f:
            tokens = [Token(t) for t in tokenize.generate_tokens(f.readline)]
        docstring_tokens = {t.string for t in get_docstring_tokens(tokens)}
        self.assertEqual(docstring_tokens, {
            "'''\nSingle quotes multiline module docstring\n'''",
            "'''\n    Single quotes multiline class docstring\n    '''",
            "'''\n        Single quotes multiline function docstring\n        '''",
        })



import tokenize

# I don't think this is a minimized state machine, but it's clearer this
# way. Namely, the class vs. function states can be merged

# In the start of the module when we're expecting possibly a string that gets marked as a docstring
STATE_EXPECT_MODULE_DOCSTRING = 0
# After seeing the class keyword, we're waiting for the block colon (and do bracket counting)
STATE_EXPECT_CLASS_COLON = 1
# After seeing the colon in a class definition we're expecting possibly a docstring
STATE_EXPECT_CLASS_DOCSTRING = 2
# Same as EXPECT_CLASS_COLON, but for function definitions
STATE_EXPECT_FUNCTION_COLON = 3
# Same as EXPECT_CLASS_DOCSTRING, but for function definitions
STATE_EXPECT_FUNCTION_DOCSTRING = 4
# Just skipping tokens until we observe a class or a def.
STATE_OTHER = 5

# These tokens don't matter here - they don't get in the way of docstrings
TOKENS_TO_IGNORE = [
    tokenize.NEWLINE,
    tokenize.INDENT,
    tokenize.DEDENT,
    tokenize.NL,
    tokenize.COMMENT,
]


def get_docstring_tokens(tokens):
    state = STATE_EXPECT_MODULE_DOCSTRING
    # The number of currently open parentheses, square brackets, etc.
    # This doesn't check if they're properly balanced, i.e. there isn't ([)], but we shouldn't
    # need to - if they aren't, it shouldn't parse at all, so we ignore the bracket type
    bracket_count = 0
    docstring_tokens = set()

    for token in tokens:
        if token.type in TOKENS_TO_IGNORE:
            continue
        if token.type == tokenize.STRING:
            if state in [STATE_EXPECT_MODULE_DOCSTRING, STATE_EXPECT_CLASS_DOCSTRING,
                         STATE_EXPECT_FUNCTION_DOCSTRING]:
                docstring_tokens.add(token)
                state = STATE_OTHER
        # A class means we'll expect the class token
        elif token.type == tokenize.NAME and token.string == 'class':
            state = STATE_EXPECT_CLASS_COLON
            # Just in case - they should be balanced normally
            bracket_count = 0
        # A def means we'll expect a colon after that
        elif token.type == tokenize.NAME and token.string == 'def':
            state = STATE_EXPECT_FUNCTION_COLON
            # Just in case - they should be balanced normally
            bracket_count = 0
        # If we get a colon and we're expecting it, move to the next state
        elif token.type == tokenize.OP and token.string == ':':
            # If there are still left brackets open, it must be something other than the block start
            if bracket_count == 0:
                if state == STATE_EXPECT_CLASS_COLON:
                    state = STATE_EXPECT_CLASS_DOCSTRING
                elif state == STATE_EXPECT_FUNCTION_COLON:
                    state = STATE_EXPECT_FUNCTION_DOCSTRING
        # Count opening and closing brackets in bracket_count
        elif token.type == tokenize.OP and token.string in ['(', '[', '{']:
            bracket_count += 1
        elif token.type == tokenize.OP and token.string in [')', ']', '}']:
            bracket_count -= 1
        # The token is not one of the recognized types. If we're expecting a colon, then all good,
        # but if we're expecting a docstring, it would no longer be a docstring
        elif state in [STATE_EXPECT_MODULE_DOCSTRING, STATE_EXPECT_CLASS_DOCSTRING,
                       STATE_EXPECT_FUNCTION_DOCSTRING]:
            state = STATE_OTHER

    return docstring_tokens




var0 = True
l = []

if var0:
    """ not a  docstring"""
    pass

while(var0 < 0 or "def" in l[:] ):
    """ also not a docstring """
    with open(l["def":]) as f:
        """ not a docstring """
        pass

if var0 < 10:
    """
        not a multiline docstring
    """
    pass


if var0:
    ''' not a  docstring'''
    pass

while(var0 < 0 or "def" in l[:] ):
    ''' also not a docstring '''
    with open(l["def":]) as f:
        ''' not a docstring '''
        pass

if var0 < 10:
    '''
        not a multiline docstring
    '''
    pass